#include <genesis.h>

#define nScreenWidth 32 // Console Screen Size X (columns)
#define nScreenHeight 28 // Console Screen Size Y (rows)
#define nMapWidth 16 // World Dimensions
#define nMapHeight 16

fix32 fPlayerY = FIX32(nMapHeight - 14.7f); // Player Start Position
fix32 fPlayerX = FIX32(5.09f);
fix32 fPlayerA = FIX32(0.0f); // Player Start Rotation
#define fFOV   (1024 * 70 / 360) // 70 градусов //FIX32(3.14159f / 16.0f) // / 4 Field of View
#define fDepth FIX32(16.0f) // Maximum rendering distance
#define fSpeed FIX32(5.0f) // Walking Speed

#define fStepSize FIX32(0.1f) // 0.1

//typedef struct
//{
//    fix32 first;
//    fix32 second;
//} pairf;

// Create Map of world space # = wall block, . = space
char map[] =
"#########......."
"#..............."
"#.......########"
"#..............#"
"#......##......#"
"#......##......#"
"#..............#"
"###............#"
"##.............#"
"#......####..###"
"#......#.......#"
"#......#.......#"
"#..............#"
"#......#########"
"#..............#"
"################";

u16 screen[32 * 28];

//fix32 acos(fix32 x)
//{
//    return FIX32(1.57079) - fix32Mul(FIX32(1.57079), x);
//}

static void handleInput()
{
    u16 value = JOY_readJoypad(JOY_1);

    fix32 fElapsedTime = FIX32(1.0/60.0);

    // Handle CCW Rotation
    if (value & BUTTON_LEFT)
        fPlayerA += fix32Mul(fix32Mul(fSpeed, FIX32(0.5f)), fElapsedTime);
    // Handle CW Rotation
    else if (value & BUTTON_RIGHT)
        fPlayerA -= fix32Mul(fix32Mul(fSpeed, FIX32(0.5f)), fElapsedTime);
    // TODO: fPlayerA в пределах 0 - 1024

    // Handle Forwards movement & collision
    if (value & BUTTON_UP)
    {
        fPlayerX += fix32Mul(fix32Mul(cosFix32(fPlayerA), fSpeed), fElapsedTime);
        fPlayerY += fix32Mul(fix32Mul(sinFix32(fPlayerA), fSpeed), fElapsedTime);
        if (map[(nMapHeight-1-fix32ToInt(fPlayerY))* nMapWidth + fix32ToInt(fPlayerX)] == '#')
        {
            fPlayerX -= fix32Mul(fix32Mul(cosFix32(fPlayerA), fSpeed), fElapsedTime);
            fPlayerY -= fix32Mul(fix32Mul(sinFix32(fPlayerA), fSpeed), fElapsedTime);
        }
    }
    // Handle backwards movement & collision
    else if (value & BUTTON_DOWN)
    {
        fPlayerX -= fix32Mul(fix32Mul(cosFix32(fPlayerA), fSpeed), fElapsedTime);
        fPlayerY -= fix32Mul(fix32Mul(sinFix32(fPlayerA), fSpeed), fElapsedTime);
        if (map[(nMapHeight-1-fix32ToInt(fPlayerY)) * nMapWidth + fix32ToInt(fPlayerX)] == '#')
        {
            fPlayerX += fix32Mul(fix32Mul(cosFix32(fPlayerA), fSpeed), fElapsedTime);
            fPlayerY += fix32Mul(fix32Mul(sinFix32(fPlayerA), fSpeed), fElapsedTime);
        }
    }
}

int main()
{
    SYS_disableInts();

    VDP_setScreenWidth256();
    VDP_setScreenHeight224();
    VDP_setPlanSize(32, 32);

    SYS_enableInts();

    while (TRUE)
    {
        VDP_showFPS(TRUE);
        //u16 rand = random() % 16;
        //for (int i = 0; i < 32 * 28; i++) {
        //    u16 tileIndex = (rand + i) % 16;
        //    screen[i] = TILE_ATTR_FULL(PAL1, FALSE, FALSE, FALSE, tileIndex);
        //}

        handleInput();

        for (s16 x = 0; x < nScreenWidth; x++)
        {
            // For each column, calculate the projected ray angle into world space
            fix32 fRayAngle = (fPlayerA + fFOV / 2) - x * fFOV / nScreenWidth;

            // Find distance to wall
            //fix32 fStepSize = FIX32(0.1f);       // Increment size for ray casting, decrease to increase
            fix32 fDistanceToWall = FIX32(0.0f); //   resolution

            u16 bHitWall = FALSE;  // Set when ray hits wall block
            //u16 bBoundary = FALSE; // Set when ray hits boundary between two wall blocks

            fix32 fEyeX = cosFix32(fRayAngle); // Unit vector for ray in player space
            fix32 fEyeY = sinFix32(fRayAngle);

            // Incrementally cast ray from player, along ray angle, testing for 
            // intersection with a block
            fix32 nextPointX = fPlayerX;
            fix32 nextPointY = fPlayerY;
            fix32 stepX = fix32Mul(fEyeX, fStepSize);
            fix32 stepY = fix32Mul(fEyeY, fStepSize);
            while (!bHitWall)// && fDistanceToWall < fDepth)
            {
                //fDistanceToWall += fStepSize;
                nextPointX += stepX;
                nextPointY += stepY;
                s32 nTestX = fix32ToInt(nextPointX);//fix32ToInt(fPlayerX + fix32Mul(fEyeX, fDistanceToWall));
                s32 nTestY = fix32ToInt(nextPointY);//fix32ToInt(fPlayerY + fix32Mul(fEyeY, fDistanceToWall));

                // Test if ray is out of bounds
                if (nTestX < 0 || nTestX >= nMapWidth || nTestY < 0 || nTestY >= nMapHeight)
                {
                    bHitWall = TRUE; // Just set distance to maximum depth
                    fDistanceToWall = fDepth;
                }
                else
                {
                    // Ray is inbounds so test to see if the ray cell is a wall block
                    if (map[(nMapHeight - 1 - nTestY) * nMapWidth + nTestX] == '#')
                    {
                        // Ray has hit wall
                        bHitWall = TRUE;

                        // Корректировка от "рыбьего глаза"
                        fDistanceToWall = fix32Mul((nextPointX - fPlayerX), cosFix32(1024 - fPlayerA))
                                        - fix32Mul((nextPointY - fPlayerY), sinFix32(1024 - fPlayerA));
                        if (fDistanceToWall < FIX32(0.1f)) {
                            fDistanceToWall = FIX32(0.1f);
                        }

                        // To highlight tile boundaries, cast a ray from each corner
                        // of the tile, to the player. The more coincident this ray
                        // is to the rendering ray, the closer we are to a tile 
                        // boundary, which we'll shade to add detail to the walls
                        //vector<pair<float, float>> p;
                        /*pairf p[4];

                        // Test each corner of hit tile, storing the distance from
                        // the player, and the calculated dot product of the two rays
                        int i = 0;
                        for (int tx = 0; tx < 2; tx++)
                            for (int ty = 0; ty < 2; ty++)
                            {
                                // Angle of corner to eye
                                fix32 vy = intToFix32(nTestY + ty) - fPlayerY;
                                fix32 vx = intToFix32(nTestX + tx) - fPlayerX;
                                fix32 d = getApproximatedDistance(vx, vy);
                                fix32 dot = fix32Mul(fEyeX, vx) + fix32Mul(fEyeY, vy);
                                p[i].first = d;
                                p[i].second = fix32Div(dot, d);
                                i++;
                            }

                        // Sort Pairs from closest to farthest
                        //sort(&p[0], &p[4], [](const pair<float, float> &left, const pair<float, float> &right) {return left.first < right.first; });

                        int iMin = 0;
                        for (int i = 0; i < 4 - 1; i++) {
                            for (int j = i + 1; j < 4; j++) {
                                if (p[iMin].first > p[j].first) {
                                    iMin = j;
                                }
                            }
                            pairf tmp = p[iMin];
                            p[iMin] = p[i];
                            p[i] = tmp;
                        }

                        // First two/three are closest (we will never see all four)
                        fix32 fBound = FIX32(0.01f);
                        if (acos(p[0].second) < fBound) bBoundary = TRUE;
                        if (acos(p[1].second) < fBound) bBoundary = TRUE;
                        if (acos(p[2].second) < fBound) bBoundary = TRUE;
                        */
                    }
                }
            }

            // Calculate distance to ceiling and floor
            s32 nCeiling = nScreenHeight / 2 - fix32ToInt(fix32Div(intToFix32(nScreenHeight), fDistanceToWall));
            s32 nFloor = nScreenHeight - nCeiling;

            // Shader walls based on distance
            u16 nShade = TILE_ATTR_FULL(PAL1, FALSE, FALSE, FALSE, 0);
            u16 tileIndex = 8 - fix32ToInt(fix32Div(min(fDistanceToWall, fDepth), fDepth) << 3);
            nShade = TILE_ATTR_FULL(PAL1, FALSE, FALSE, FALSE, tileIndex);
            //if (fDistanceToWall <= fDepth / 4)     nShade = TILE_ATTR_FULL(PAL1, FALSE, FALSE, FALSE, 8); // Very close
            //else if (fDistanceToWall < fDepth / 3) nShade = TILE_ATTR_FULL(PAL1, FALSE, FALSE, FALSE, 6);
            //else if (fDistanceToWall < fDepth / 2) nShade = TILE_ATTR_FULL(PAL1, FALSE, FALSE, FALSE, 4);
            //else if (fDistanceToWall < fDepth)     nShade = TILE_ATTR_FULL(PAL1, FALSE, FALSE, FALSE, 2);
            //else                                   nShade = TILE_ATTR_FULL(PAL1, FALSE, FALSE, FALSE, 0); // Too far away

            //if (bBoundary) nShade = TILE_ATTR_FULL(PAL3, FALSE, FALSE, FALSE, 9); // Black it out

            for (s16 y = 0; y < nScreenHeight; y++)
            {
                // Each Row
                if (y <= nCeiling)
                    screen[y*nScreenWidth + x] = TILE_ATTR_FULL(PAL1, FALSE, FALSE, FALSE, 0);
                else if (y > nCeiling && y <= nFloor)
                    screen[y*nScreenWidth + x] = nShade;
                else {
                    screen[y*nScreenWidth + x] = TILE_ATTR_FULL(PAL1, FALSE, FALSE, FALSE, 0);
                }
                /*else // Floor
                {
                    // Shade floor based on distance
                    float b = 1.0f - (((float)y - nScreenHeight / 2) / ((float)nScreenHeight / 2));
                    if (b < 0.25)		nShade = '#';
                    else if (b < 0.5)	nShade = 'x';
                    else if (b < 0.75)	nShade = '.';
                    else if (b < 0.9)	nShade = '-';
                    else				nShade = ' ';
                    screen[y*nScreenWidth + x] = nShade;
                }*/
            }
        }

        // Display Stats
        //swprintf_s(screen, 40, L"X=%3.2f, Y=%3.2f, A=%3.2f FPS=%3.2f ", fPlayerX, fPlayerY, fPlayerA, 1.0f / fElapsedTime);

        // Display Map
        //for (int nx = 0; nx < nMapWidth; nx++)
        //    for (int ny = 0; ny < nMapWidth; ny++)
        //    {
        //        screen[(ny + 1)*nScreenWidth + nx] = map[ny * nMapWidth + nx];
        //    }
        //screen[((int)fPlayerX + 1) * nScreenWidth + (int)fPlayerY] = 'P';

        // Display Frame
        //screen[nScreenWidth * nScreenHeight - 1] = '\0';
        //WriteConsoleOutputCharacter(hConsole, screen, nScreenWidth * nScreenHeight, { 0,0 }, &dwBytesWritten);

        VDP_setTileMapData(VDP_PLAN_B, screen, 0, 32 * 28, DMA_QUEUE);
        VDP_waitVSync();
    }

    return 0;
}
